@@ -1,5 +1,5 @@
 /*
-    $Id: strscan.c 37916 2012-11-28 00:17:33Z ryan $
+    $Id: strscan.c 52988 2015-12-09 01:01:17Z ko1 $
 
     Copyright (c) 1999-2006 Minero Aoki
 
@@ -38,6 +38,9 @@ struct strscanner
 
     /* the regexp register; legal only when MATCHED_P(s) */
     struct re_registers regs;
+
+    /* regexp used for last scan */
+    VALUE regex;
 };
 
 #define MATCHED_P(s)          ((s)->flags & FLAG_MATCHED)
@@ -178,29 +181,25 @@ static size_t
 strscan_memsize(const void *ptr)
 {
     const struct strscanner *p = ptr;
-    size_t size = 0;
-    if (p) {
-	size = sizeof(*p) - sizeof(p->regs) + onig_region_memsize(&p->regs);
-    }
-    return size;
+    return sizeof(*p) - sizeof(p->regs) + onig_region_memsize(&p->regs);
 }
 
 static const rb_data_type_t strscanner_type = {
     "StringScanner",
-    {strscan_mark, strscan_free, strscan_memsize}
+    {strscan_mark, strscan_free, strscan_memsize},
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static VALUE
 strscan_s_allocate(VALUE klass)
 {
     struct strscanner *p;
+    VALUE obj = TypedData_Make_Struct(klass, struct strscanner, &strscanner_type, p);
 
-    p = ALLOC(struct strscanner);
-    MEMZERO(p, struct strscanner, 1);
     CLEAR_MATCH_STATUS(p);
     onig_region_init(&(p->regs));
     p->str = Qnil;
-    return TypedData_Wrap_Struct(klass, &strscanner_type, p);
+    return obj;
 }
 
 /*
@@ -248,7 +247,9 @@ strscan_init_copy(VALUE vself, VALUE vorig)
 	self->str = orig->str;
 	self->prev = orig->prev;
 	self->curr = orig->curr;
-	onig_region_copy(&self->regs, &orig->regs);
+	if (rb_reg_region_copy(&self->regs, &orig->regs))
+	    rb_memerror();
+	RB_GC_GUARD(vorig);
     }
 
     return vself;
@@ -456,6 +457,8 @@ strscan_do_scan(VALUE self, VALUE regex, int succptr, int getstr, int headonly)
     if (S_RESTLEN(p) < 0) {
         return Qnil;
     }
+
+    p->regex = regex;
     re = rb_reg_prepare_re(regex, p->str);
     tmpreg = re != RREGEXP(regex)->ptr;
     if (!tmpreg) RREGEXP(regex)->usecnt++;
@@ -970,6 +973,24 @@ strscan_matched_size(VALUE self)
     return INT2NUM(p->regs.end[0] - p->regs.beg[0]);
 }
 
+static int
+name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name, const char* name_end, rb_encoding *enc)
+{
+    int num;
+
+    num = onig_name_to_backref_number(RREGEXP(regexp)->ptr,
+	(const unsigned char* )name, (const unsigned char* )name_end, regs);
+    if (num >= 1) {
+	return num;
+    }
+    else {
+	rb_enc_raise(enc, rb_eIndexError, "undefined group name reference: %.*s",
+					  rb_long2int(name_end - name), name);
+    }
+
+    UNREACHABLE;
+}
+
 /*
  * call-seq: [](n)
  *
@@ -983,17 +1004,41 @@ strscan_matched_size(VALUE self)
  *   s[3]                               # -> "12"
  *   s.post_match                       # -> "1975 14:39"
  *   s.pre_match                        # -> ""
+ *
+ *   s.reset
+ *   s.scan(/(?<wday>\w+) (?<month>\w+) (?<day>\d+) /)       # -> "Fri Dec 12 "
+ *   s[0]                               # -> "Fri Dec 12 "
+ *   s[1]                               # -> "Fri"
+ *   s[2]                               # -> "Dec"
+ *   s[3]                               # -> "12"
+ *   s[:wday]                           # -> "Fri"
+ *   s[:month]                          # -> "Dec"
+ *   s[:day]                            # -> "12"
+ *   s.post_match                       # -> "1975 14:39"
+ *   s.pre_match                        # -> ""
  */
 static VALUE
 strscan_aref(VALUE self, VALUE idx)
 {
+    const char *name;
     struct strscanner *p;
     long i;
 
     GET_SCANNER(self, p);
     if (! MATCHED_P(p))        return Qnil;
 
-    i = NUM2LONG(idx);
+    switch (TYPE(idx)) {
+        case T_SYMBOL:
+            idx = rb_sym2str(idx);
+            /* fall through */
+        case T_STRING:
+            RSTRING_GETMEM(idx, name, i);
+            i = name_to_backref_number(&(p->regs), p->regex, name, name + i, rb_enc_get(idx));
+            break;
+        default:
+            i = NUM2LONG(idx);
+    }
+
     if (i < 0)
         i += p->regs.num_regs;
     if (i < 0)                 return Qnil;
@@ -1087,7 +1132,6 @@ strscan_restsize(VALUE self)
 }
 
 #define INSPECT_LENGTH 5
-#define BUFSIZE 256
 
 /*
  * Returns a string that represents the StringScanner object, showing:
@@ -1104,76 +1148,69 @@ static VALUE
 strscan_inspect(VALUE self)
 {
     struct strscanner *p;
-    char buf[BUFSIZE];
-    long len;
     VALUE a, b;
 
     p = check_strscan(self);
     if (NIL_P(p->str)) {
-        len = snprintf(buf, BUFSIZE, "#<%s (uninitialized)>",
-                       rb_class2name(CLASS_OF(self)));
-        return infect(rb_str_new(buf, len), p);
+	a = rb_sprintf("#<%"PRIsVALUE" (uninitialized)>", rb_obj_class(self));
+	return infect(a, p);
     }
     if (EOS_P(p)) {
-        len = snprintf(buf, BUFSIZE, "#<%s fin>",
-                       rb_class2name(CLASS_OF(self)));
-        return infect(rb_str_new(buf, len), p);
+	a = rb_sprintf("#<%"PRIsVALUE" fin>", rb_obj_class(self));
+	return infect(a, p);
     }
     if (p->curr == 0) {
-        b = inspect2(p);
-        len = snprintf(buf, BUFSIZE, "#<%s %ld/%ld @ %s>",
-                       rb_class2name(CLASS_OF(self)),
-                       p->curr, S_LEN(p),
-                       RSTRING_PTR(b));
-        return infect(rb_str_new(buf, len), p);
+	b = inspect2(p);
+	a = rb_sprintf("#<%"PRIsVALUE" %ld/%ld @ %"PRIsVALUE">",
+		       rb_obj_class(self),
+		       p->curr, S_LEN(p),
+		       b);
+	return infect(a, p);
     }
     a = inspect1(p);
     b = inspect2(p);
-    len = snprintf(buf, BUFSIZE, "#<%s %ld/%ld %s @ %s>",
-                   rb_class2name(CLASS_OF(self)),
-                   p->curr, S_LEN(p),
-                   RSTRING_PTR(a),
-                   RSTRING_PTR(b));
-    return infect(rb_str_new(buf, len), p);
+    a = rb_sprintf("#<%"PRIsVALUE" %ld/%ld %"PRIsVALUE" @ %"PRIsVALUE">",
+		   rb_obj_class(self),
+		   p->curr, S_LEN(p),
+		   a, b);
+    return infect(a, p);
 }
 
 static VALUE
 inspect1(struct strscanner *p)
 {
-    char buf[BUFSIZE];
-    char *bp = buf;
+    VALUE str;
     long len;
 
     if (p->curr == 0) return rb_str_new2("");
     if (p->curr > INSPECT_LENGTH) {
-        strcpy(bp, "..."); bp += 3;
-        len = INSPECT_LENGTH;
+	str = rb_str_new_cstr("...");
+	len = INSPECT_LENGTH;
     }
     else {
-        len = p->curr;
+	str = rb_str_new(0, 0);
+	len = p->curr;
     }
-    memcpy(bp, CURPTR(p) - len, len); bp += len;
-    return rb_str_dump(rb_str_new(buf, bp - buf));
+    rb_str_cat(str, CURPTR(p) - len, len);
+    return rb_str_dump(str);
 }
 
 static VALUE
 inspect2(struct strscanner *p)
 {
-    char buf[BUFSIZE];
-    char *bp = buf;
+    VALUE str;
     long len;
 
     if (EOS_P(p)) return rb_str_new2("");
     len = S_LEN(p) - p->curr;
     if (len > INSPECT_LENGTH) {
-        len = INSPECT_LENGTH;
-        memcpy(bp, CURPTR(p), len); bp += len;
-        strcpy(bp, "..."); bp += 3;
+	str = rb_str_new(CURPTR(p), INSPECT_LENGTH);
+	rb_str_cat2(str, "...");
     }
     else {
-        memcpy(bp, CURPTR(p), len); bp += len;
+	str = rb_str_new(CURPTR(p), len);
     }
-    return rb_str_dump(rb_str_new(buf, bp - buf));
+    return rb_str_dump(str);
 }
 
 /* =======================================================================
@@ -1284,7 +1321,7 @@ inspect2(struct strscanner *p)
  * There are aliases to several of the methods.
  */
 void
-Init_strscan()
+Init_strscan(void)
 {
     ID id_scanerr = rb_intern("ScanError");
     VALUE tmp;
@@ -1299,7 +1336,7 @@ Init_strscan()
     tmp = rb_str_new2(STRSCAN_VERSION);
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Version"), tmp);
-    tmp = rb_str_new2("$Id: strscan.c 37916 2012-11-28 00:17:33Z ryan $");
+    tmp = rb_str_new2("$Id: strscan.c 52988 2015-12-09 01:01:17Z ko1 $");
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Id"), tmp);
 
