@@ -2,7 +2,7 @@
 
   insnhelper.h - helper macros to implement each instructions
 
-  $Author: naruse $
+  $Author: ko1 $
   created at: 04/01/01 15:50:34 JST
 
   Copyright (C) 2004-2007 Koichi Sasada
@@ -12,54 +12,6 @@
 #ifndef RUBY_INSNHELPER_H
 #define RUBY_INSNHELPER_H
 
-/**
- * VM Debug Level
- *
- * debug level:
- *  0: no debug output
- *  1: show instruction name
- *  2: show stack frame when control stack frame is changed
- *  3: show stack status
- *  4: show register
- *  5:
- * 10: gc check
- */
-
-#ifndef VMDEBUG
-#define VMDEBUG 0
-#endif
-
-#if 0
-#undef  VMDEBUG
-#define VMDEBUG 3
-#endif
-
-enum {
-  BOP_PLUS,
-  BOP_MINUS,
-  BOP_MULT,
-  BOP_DIV,
-  BOP_MOD,
-  BOP_EQ,
-  BOP_EQQ,
-  BOP_LT,
-  BOP_LE,
-  BOP_LTLT,
-  BOP_AREF,
-  BOP_ASET,
-  BOP_LENGTH,
-  BOP_SIZE,
-  BOP_EMPTY_P,
-  BOP_SUCC,
-  BOP_GT,
-  BOP_GE,
-  BOP_NOT,
-  BOP_NEQ,
-
-  BOP_LAST_
-};
-
-extern char ruby_vm_redefined_flag[BOP_LAST_];
 extern VALUE ruby_vm_const_missing_count;
 
 #if VM_COLLECT_USAGE_DETAILS
@@ -127,8 +79,6 @@ enum vm_regan_acttype {
 #define GET_CURRENT_INSN() (*GET_PC())
 #define GET_OPERAND(n)     (GET_PC()[(n)])
 #define ADD_PC(n)          (SET_PC(REG_PC + (n)))
-
-#define GET_PC_COUNT()     (REG_PC - GET_ISEQ()->iseq_encoded)
 #define JUMP(dst)          (REG_PC += (dst))
 
 /* frame pointer, environment pointer */
@@ -171,27 +121,8 @@ enum vm_regan_acttype {
 /* deal with control flow 2: method/iterator              */
 /**********************************************************/
 
-#define COPY_CREF_OMOD(c1, c2) do {  \
-  (c1)->nd_refinements = (c2)->nd_refinements; \
-  if (!NIL_P((c2)->nd_refinements)) { \
-      (c1)->flags |= NODE_FL_CREF_OMOD_SHARED; \
-      (c2)->flags |= NODE_FL_CREF_OMOD_SHARED; \
-  } \
-} while (0)
-
-#define COPY_CREF(c1, c2) do {  \
-  NODE *__tmp_c2 = (c2); \
-  COPY_CREF_OMOD(c1, __tmp_c2); \
-  (c1)->nd_clss = __tmp_c2->nd_clss; \
-  (c1)->nd_visi = __tmp_c2->nd_visi;\
-  (c1)->nd_next = __tmp_c2->nd_next; \
-  if (__tmp_c2->flags & NODE_FL_CREF_PUSHED_BY_EVAL) { \
-      (c1)->flags |= NODE_FL_CREF_PUSHED_BY_EVAL; \
-  } \
-} while (0)
-
-#define CALL_METHOD(ci) do { \
-    VALUE v = (*(ci)->call)(th, GET_CFP(), (ci)); \
+#define CALL_METHOD(calling, ci, cc) do { \
+    VALUE v = (*(cc)->call)(th, GET_CFP(), (calling), (ci), (cc)); \
     if (v == Qundef) { \
 	RESTORE_REGS(); \
 	NEXT_INSN(); \
@@ -210,8 +141,8 @@ enum vm_regan_acttype {
 #endif
 
 #if OPT_CALL_FASTPATH
-#define CI_SET_FASTPATH(ci, func, enabled) do { \
-    if (LIKELY(enabled)) ((ci)->call = (func)); \
+#define CI_SET_FASTPATH(cc, func, enabled) do { \
+    if (LIKELY(enabled)) ((cc)->call = (func)); \
 } while (0)
 #else
 #define CI_SET_FASTPATH(ci, func, enabled) /* do nothing */
@@ -229,46 +160,68 @@ enum vm_regan_acttype {
 /**********************************************************/
 
 /* optimize insn */
-#define FIXNUM_REDEFINED_OP_FLAG (1 << 0)
-#define FLOAT_REDEFINED_OP_FLAG  (1 << 1)
-#define STRING_REDEFINED_OP_FLAG (1 << 2)
-#define ARRAY_REDEFINED_OP_FLAG  (1 << 3)
-#define HASH_REDEFINED_OP_FLAG   (1 << 4)
-#define BIGNUM_REDEFINED_OP_FLAG (1 << 5)
-#define SYMBOL_REDEFINED_OP_FLAG (1 << 6)
-#define TIME_REDEFINED_OP_FLAG   (1 << 7)
-
-#define BASIC_OP_UNREDEFINED_P(op, klass) (LIKELY((ruby_vm_redefined_flag[(op)]&(klass)) == 0))
-
 #define FIXNUM_2_P(a, b) ((a) & (b) & 1)
 #if USE_FLONUM
 #define FLONUM_2_P(a, b) (((((a)^2) | ((b)^2)) & 3) == 0) /* (FLONUM_P(a) && FLONUM_P(b)) */
 #else
 #define FLONUM_2_P(a, b) 0
 #endif
-#define HEAP_CLASS_OF(obj) (RBASIC(obj)->klass)
 
 #ifndef USE_IC_FOR_SPECIALIZED_METHOD
 #define USE_IC_FOR_SPECIALIZED_METHOD 1
 #endif
 
-#define CALL_SIMPLE_METHOD(recv) do { \
-    ci->blockptr = 0; ci->argc = ci->orig_argc; \
-    vm_search_method(ci, ci->recv = (recv)); \
-    CALL_METHOD(ci); \
+#define CALL_SIMPLE_METHOD(recv_) do { \
+    struct rb_calling_info calling; \
+    calling.blockptr = NULL; \
+    calling.argc = ci->orig_argc; \
+    vm_search_method(ci, cc, calling.recv = (recv_)); \
+    CALL_METHOD(&calling, ci, cc); \
 } while (0)
 
-static VALUE ruby_vm_global_state_version = 1;
-
-#define GET_VM_STATE_VERSION() (ruby_vm_global_state_version)
-#define INC_VM_STATE_VERSION() do { \
-    ruby_vm_global_state_version = (ruby_vm_global_state_version + 1); \
-    if (ruby_vm_global_state_version == 0) vm_clear_all_cache(); \
-} while (0)
-static void vm_clear_all_cache(void);
+#define NEXT_CLASS_SERIAL() (++ruby_vm_class_serial)
+#define GET_GLOBAL_METHOD_STATE() (ruby_vm_global_method_state)
+#define INC_GLOBAL_METHOD_STATE() (++ruby_vm_global_method_state)
+#define GET_GLOBAL_CONSTANT_STATE() (ruby_vm_global_constant_state)
+#define INC_GLOBAL_CONSTANT_STATE() (++ruby_vm_global_constant_state)
 
 static VALUE make_no_method_exception(VALUE exc, const char *format,
 				      VALUE obj, int argc, const VALUE *argv);
 
+static inline struct vm_throw_data *
+THROW_DATA_NEW(VALUE val, rb_control_frame_t *cf, VALUE st)
+{
+    return (struct vm_throw_data *)rb_imemo_new(imemo_throw_data, val, (VALUE)cf, st, 0);
+}
+
+static inline void
+THROW_DATA_CATCH_FRAME_SET(struct vm_throw_data *obj, const rb_control_frame_t *cfp)
+{
+    obj->catch_frame = cfp;
+}
+
+static inline void
+THROW_DATA_STATE_SET(struct vm_throw_data *obj, int st)
+{
+    obj->throw_state = (VALUE)st;
+}
+
+static inline VALUE
+THROW_DATA_VAL(const struct vm_throw_data *obj)
+{
+    return obj->throw_obj;
+}
+
+static inline const rb_control_frame_t *
+THROW_DATA_CATCH_FRAME(const struct vm_throw_data *obj)
+{
+    return obj->catch_frame;
+}
+
+static int
+THROW_DATA_STATE(const struct vm_throw_data *obj)
+{
+    return (int)obj->throw_state;
+}
 
 #endif /* RUBY_INSNHELPER_H */
